@@ -21,10 +21,10 @@
 __author__ = """Peter Rowlands"""
 __email__ = 'peter@pmrowla.com'
 __license__ = 'GPLv3'
-__version__ = '0.1.0'
+__version__ = '0.2.2'
 
 
-from .archive import LMArchive, LMArchiveInfo
+from .archive import LMArchive, LMArchiveInfo, LMCompressType
 from .lsb import LMScript, LNSCompiler, LNSDecompiler
 
-__all__ = ['LMArchive', 'LMArchiveInfo', 'LMScript', 'LNSCompiler', 'LNSDecompiler']
+__all__ = ['LMArchive', 'LMArchiveInfo', 'LMCompressType', 'LMScript', 'LNSCompiler', 'LNSDecompiler']
