@@ -25,15 +25,11 @@ The API for `archive` behaves similary to Python's ``zipfile`` module,
 with the exception that archives cannot be modified in-place (i.e. mode ``'a'``
 is unavailable).
 
-Note:
-    Reading (and writing) files which are encrypted via LiveMaker Pro is not
-    supported, but any non-encrypted files from a LiveMaker Pro archive can
-    be extracted.
-
 """
 
 import enum
 import logging
+import os.path
 import shutil
 import struct
 import tempfile
@@ -44,6 +40,7 @@ from pathlib import Path, PureWindowsPath
 import construct
 
 from .exceptions import BadLiveMakerArchive, UnsupportedLiveMakerCompression
+from .scramble import decrypt
 
 
 log = logging.getLogger(__name__)
@@ -59,7 +56,9 @@ class LMCompressType(enum.IntEnum):
 
 SUPPORTED_COMPRESSIONS = [
     LMCompressType.NONE,
-    LMCompressType.ZLIB
+    LMCompressType.ZLIB,
+    LMCompressType.ENCRYPTED,
+    LMCompressType.ENCRYPTED_ZLIB,
 ]
 
 # LM3 seed for TpRandom
@@ -136,6 +135,9 @@ VF_CHECKSUM_KEYS = [
 MIN_ARCHIVE_VERSION = 100
 DEFAULT_VERSION = 102
 
+# Max size of one split archive part (1GB)
+SPLIT_ARCHIVE_PART_SIZE = 1073741824
+
 
 class LMObfuscator(object):
     """Class for (de)obfuscating LiveMaker directory fields.
@@ -369,7 +371,30 @@ class LMArchive(object):
     Behaves in the same manner as Python ``tarfile.TarFile`` or ``zipfile.ZipFile``.
     Can be used inside a Python ``with`` statement (in the same way as zip/tar files).
 
+    Args:
+        name: Pathname for the archive. `name` can be a string or path-like object.
+            If omitted, `fp` must be specified.
+        mode: Either ``'r'`` to read from an existing archive or ``'w'`` to create a new file
+            (overwriting an existing one). Defaults to ``'r'``.
+        fp: If `fileobj` is given, it will be used for reading and writing data.
+            If it can be determined, `mode` will be overridden by `fp`'s mode. `fileobj`
+            will be used from position 0.
+        exe: Pathname for LiveMaker executable (``.exe``) file. If `exe` is given and
+            `mode` is ``w``, the output file will be an executable with the archive
+            appended to the end (i.e. a LiveMaker ``.exe`` file). If `exe` is not given,
+            the output file will be a standalone archive (i.e. a LiveMaker ``.dat`` file).
+            `exe` does nothing when opening a file for reading.
+        split: If opening in write mode and `split` is ``True``, the archive data will be split
+            into smaller files across 1GB boundaries. `split` has no effect in read mode or when
+            writing an executable archive.
+        version: Archive version, only applies to write mode.
+
     Note:
+        `fp` is not closed when `LiveMakerArchive` is closed.
+
+        ``'a'`` is an invalid `mode` for `LiveMakerArchive`. Archives cannot be modified
+        in place, to patch an existing archive, you must write to a new file.
+
         When opened in write mode, the output archive file will not be written until
         `LMArchive.close()` is called. As entries are added to the archive, they will
         be written to a temporary file. Upon calling `close()`, the exe (if it exists)
@@ -378,33 +403,9 @@ class LMArchive(object):
 
     """
 
-    def __init__(self, name=None, mode='r', fp=None, exe=None, version=DEFAULT_VERSION):
-        """Open LiveMaker archive `name` for reading or writing.
-
-        Args:
-            name: Pathname for the archive. `name` can be a string or path-like object.
-                If omitted, `fp` must be specified.
-            mode: Either ``'r'`` to read from an existing archive or ``'w'`` to create a new file
-                (overwriting an existing one). Defaults to ``'r'``.
-            fp: If `fileobj` is given, it will be used for reading and writing data.
-                If it can be determined, `mode` will be overridden by `fp`'s mode. `fileobj`
-                will be used from position 0.
-            exe: Pathname for LiveMaker executable (``.exe``) file. If `exe` is given and
-                `mode` is ``w``, the output file will be an executable with the archive
-                appended to the end (i.e. a LiveMaker ``.exe`` file). If `exe` is not given,
-                the output file will be a standalone archive (i.e. a LiveMaker ``.dat`` file).
-                `exe` does nothing when opening a file for reading.
-            version: Archive version, only applies to write mode.
-
-        Note:
-            `fp` is not closed when `LiveMakerArchive` is closed.
-
-            ``'a'`` is an invalid `mode` for `LiveMakerArchive`. Archives cannot be modified
-            in place, to patch an existing archive, you must write to a new file.
-
-        """
+    def __init__(self, name=None, mode='r', fp=None, exe=None, split=False, version=DEFAULT_VERSION):
         self.closed = True
-        if not name and fp:
+        if not name and not fp:
             raise ValueError('Nothing to open')
         modes = {'r': 'rb', 'w': 'wb'}
         if mode not in modes:
@@ -424,21 +425,35 @@ class LMArchive(object):
         self.fp = fp
         self.closed = False
 
-        if name:
-            self.name = str(Path(name).resolve())
-        else:
-            self.name = None
+        self.name = str(Path(name).resolve())
+        root, ext = os.path.splitext(self.name)
 
         self.filelist = []
         self.name_info = {}
 
         try:
             if mode == 'w':
+                self._read_fps = []
                 if exe:
                     self.exefp = open(exe, 'rb')
                     self.is_exe = True
+                    self.is_split = False
+                    self.has_ext = False
                 else:
                     self.is_exe = False
+                    if split:
+                        self.is_split = True
+                        self._split_base = root
+                        self._split_files = set([self.name])
+                        if ext.lower() not in ['.dat', '.ext']:
+                            log.warn('Writing split archive index without .dat or .ext file extension.')
+                        elif ext.lower() == '.ext':
+                            self.has_ext = True
+                        else:
+                            self.has_ext = False
+                    else:
+                        self.is_split = False
+                        self.has_ext = False
                     self.exefp = None
                 self.tmpfp = tempfile.TemporaryFile()
                 self.version = version
@@ -446,13 +461,41 @@ class LMArchive(object):
                 self.exefp = None
                 self.is_exe = False
                 self.tmpfp = None
+                self._read_fps = []
                 savepos = self.fp.tell()
                 self.archive_offset = self._find_archive_offset()
                 self._parse_directory()
+                if ext.lower() in ['.dat', '.ext']:
+                    self.is_split = True
+                    self._split_base = root
+                    self._split_files = set([self.name])
+                    if ext.lower() == '.ext':
+                        self.has_ext = True
+                        dat_file = '{}.dat'.format(self._split_base)
+                        if not os.path.isfile(dat_file):
+                            raise BadLiveMakerArchive(
+                                'Could not find (.dat) data file for split archive index {}.'.format(self.name))
+                        self._split_files.add(dat_file)
+                        self._read_fps.append(open(dat_file, self._mode))
+                    else:
+                        self.has_ext = False
+                        self._read_fps.append(self.fp)
+                    for i in range(1, 100):
+                        dat_file = '{}.{:03}'.format(self._split_base, i)
+                        if os.path.isfile(dat_file):
+                            self._read_fps.append(open(dat_file, self._mode))
+                            self._split_files.add(dat_file)
+                else:
+                    self.is_split = False
+                    self.has_ext = False
         except Exception as e:
             if self._extfp:
                 self.fp.seek(savepos)
             else:
+                if self._read_fps:
+                    for fp in self._read_fps:
+                        if fp != self.fp:
+                            fp.close()
                 self.fp.close()
             self.closed = True
             raise e
@@ -483,12 +526,16 @@ class LMArchive(object):
                 self._write_archive()
                 self._write_trailer()
         finally:
-            if not self._extfp:
-                self.fp.close()
             if self.exefp:
                 self.exefp.close()
             if self.tmpfp:
                 self.tmpfp.close()
+            if self._read_fps:
+                for fp in self._read_fps:
+                    if fp != self.fp:
+                        fp.close()
+            if not self._extfp:
+                self.fp.close()
             self.closed = True
 
     def namelist(self):
@@ -515,7 +562,7 @@ class LMArchive(object):
         """Print a list of archive entries to stdout."""
         print(self.name)
         print('-------- ------ -------- -------- ------- ------')
-        print(' Length   mode    unk1    chksum   flags   Name')
+        print(' Length   Mode    unk1    Chksum   Flags   Name')
         print('-------- ------ -------- -------- ------- ------')
         for info in self.infolist():
             print('{:8} {:6} {:08x} {:08x} {:02x} {}'.format(
@@ -652,14 +699,35 @@ class LMArchive(object):
             info = self.getinfo(name)
         if decompress and info.compress_type not in SUPPORTED_COMPRESSIONS:
             raise UnsupportedLiveMakerCompression('{} is unsupported'.format(info.compress_type))
-        self.fp.seek(self.archive_offset + info._offset)
-        data = self.fp.read(info.compressed_size)
+        if self._read_fps:
+            # archive data is split on 1GB (1024 * 1024 * 1024 bytes) boundaries
+            # start reading from whichever data file contains the start of
+            # this node, and then continue reading across boundary as needed
+            offset = self.archive_offset + info._offset
+            data = b''
+            while len(data) < info.compressed_size:
+                bytes_needed = info.compressed_size - len(data)
+                fp = self._read_fps[offset // SPLIT_ARCHIVE_PART_SIZE]
+                fp.seek(offset % SPLIT_ARCHIVE_PART_SIZE)
+                tmp = fp.read(bytes_needed)
+                data += tmp
+                offset += len(tmp)
+        else:
+            fp = self.fp
+            fp.seek(self.archive_offset + info._offset)
+            data = self.fp.read(info.compressed_size)
         if not skip_checksum and info.checksum is not None:
             if info.checksum != LMArchiveDirectory.checksum(data):
                 log.warn('Bad checksum for file {}.'.format(info.name))
         if decompress:
-            if info.compress_type == LMCompressType.ZLIB:
-                data = zlib.decompress(data)
+            if info.compress_type in (LMCompressType.ENCRYPTED, LMCompressType.ENCRYPTED_ZLIB):
+                data = decrypt(data)
+
+            if info.compress_type in (LMCompressType.ZLIB, LMCompressType.ENCRYPTED_ZLIB):
+                try:
+                    data = zlib.decompress(data)
+                except zlib.error as e:
+                    raise UnsupportedLiveMakerCompression(str(e))
         return data
 
     def read_exe(self):
@@ -679,7 +747,7 @@ class LMArchive(object):
         self.fp.seek(0)
         return self.fp.read(self.archive_offset)
 
-    def write(self, filename, arcname=None, compress_type=None):
+    def write(self, filename, arcname=None, compress_type=None, unk1=None):
         """Write the file named `filename` into the archive.
 
         Args:
@@ -724,6 +792,8 @@ class LMArchive(object):
                 compress_type = LMCompressType.NONE
         if compress_type == LMCompressType.ZLIB:
             data = zlib.compress(data)
+        if unk1 is not None:
+            info.unk1 = unk1
         info.compress_type = compress_type
         info.compressed_size = len(data)
         info.checksum = LMArchiveDirectory.checksum(data)
@@ -817,29 +887,63 @@ class LMArchive(object):
             # info offset will be the offset of this entry in the temporary
             # file (i.e. starting at 0). archive offsets need to start from the
             # end of the directory.
-            offsets.append(LMArchiveDirectory.split_offset(info._offset + directory_size))
+            offset = info._offset
+            if not self.is_split or not self.has_ext:
+                offset = info._offset + directory_size
+            offsets.append(LMArchiveDirectory.split_offset(offset))
             compress_types.append(info.compress_type)
             unk1s.append(info.unk1)
             checksums.append(info.checksum)
             encrypt_flags.append(info.encrypt_flag)
         last_entry = self.filelist[-1]
         if last_entry is not None:
-            offsets.append(LMArchiveDirectory.split_offset(
-                last_entry._offset + directory_size + last_entry.compressed_size))
+            offset = last_entry._offset + last_entry.compressed_size
+            if not self.is_split or not self.has_ext:
+                offset += directory_size
+            offsets.append(LMArchiveDirectory.split_offset(offset))
         else:
+            # handle empty archive
             offsets.append(LMArchiveDirectory.split_offset(archive_offset))
         data = LMArchiveDirectory.struct().build(directory)
         self.fp.write(data)
 
     def _write_archive(self):
         # copy data from temp file into final archive
-        self.tmpfp.seek(0)
-        shutil.copyfileobj(self.tmpfp, self.fp)
+        if self.is_split:
+            self.tmpfp.seek(0, 2)
+            data_offset = self.fp.tell()
+            if data_offset > SPLIT_ARCHIVE_PART_SIZE:
+                raise BadLiveMakerArchive('Cannot generate split archive with exe+directory size > 1GB')
+            if self.has_ext:
+                total_size = self.tmpfp.tell()
+                extra_files = total_size // SPLIT_ARCHIVE_PART_SIZE
+            else:
+                total_size = data_offset + self.tmpfp.tell()
+                extra_files = total_size // SPLIT_ARCHIVE_PART_SIZE
+            if total_size and (total_size % SPLIT_ARCHIVE_PART_SIZE) == 0:
+                extra_files -= 1
+            self.tmpfp.seek(0)
+            if self.has_ext:
+                dat_file = '{}.dat'.format(self._split_base)
+                with open(dat_file, self._mode) as fp:
+                    fp.write(self.tmpfp.read(SPLIT_ARCHIVE_PART_SIZE))
+                self._split_files.add(dat_file)
+            else:
+                self.fp.write(self.tmpfp.read(SPLIT_ARCHIVE_PART_SIZE - data_offset))
+            for i in range(1, extra_files + 1):
+                dat_file = '{}.{:03}'.format(self._split_base, i)
+                with open(dat_file, self._mode) as fp:
+                    fp.write(self.tmpfp.read(SPLIT_ARCHIVE_PART_SIZE))
+                self._split_files.add(dat_file)
+        else:
+            self.tmpfp.seek(0)
+            shutil.copyfileobj(self.tmpfp, self.fp)
 
     def _write_trailer(self):
-        data = struct.pack('<I', self.archive_offset)
-        self.fp.write(data)
-        self.fp.write(b'lv')
+        if not self.is_split:
+            data = struct.pack('<I', self.archive_offset)
+            self.fp.write(data)
+            self.fp.write(b'lv')
 
 
 class LMArchiveInfo(object):
